@@ -1,8 +1,24 @@
-use std::env;
 use std::fs;
 use std::io::{self, Write};
 use std::path::{Path, PathBuf};
 
+/// Like std::env::var, but informs cargo.
+///
+/// Always use this one instead of the function in the standard
+/// library!
+///
+/// If we look at an environment variable, we need to tell cargo that
+/// we're doing so.  This makes sure that cargo will consider the
+/// build stale if the value of the environment variable changes.
+fn env_var<K>(key: K)-> std::result::Result<String, std::env::VarError>
+where
+    K: AsRef<std::ffi::OsStr>,
+{
+    let key = key.as_ref();
+    println!("cargo:rerun-if-env-changed={}", key.to_string_lossy());
+    std::env::var(key)
+}
+
 type Result<T> = std::result::Result<T, Box<dyn std::error::Error>>;
 
 /// Build-time configuration.
@@ -12,6 +28,9 @@ struct Config {
 
     /// Cv448/Ed448 support in Nettle.
     have_cv448: bool,
+
+    /// OCB support in Nettle.
+    have_ocb: bool,
 }
 
 /// Returns true if the given Nettle version matches or exceeds our
@@ -45,7 +64,7 @@ fn is_version_at_least<V: AsRef<str>>(nettle_version: V, need: &[u8]) -> bool {
 /// Returns whether or not the feature detection should be overridden,
 /// and if so, whether the feature should be enabled.
 fn check_override(feature: &str) -> Option<bool> {
-    env::var(&format!("NETTLE_HAVE_{}", feature))
+    env_var(&format!("NETTLE_HAVE_{}", feature))
         .ok()
         .map(|v| match v.to_lowercase().as_str() {
             "1" | "y" | "yes" | "enable" | "true" => true,
@@ -88,6 +107,13 @@ fn check_cv448(includes: &[PathBuf]) -> bool {
                                     "nettle_ed448_shake256_sign"))
 }
 
+/// Checks whether Nettle supports OCB mode.
+fn check_ocb(includes: &[PathBuf]) -> bool {
+    check_override("OCB")
+        .unwrap_or_else(|| check_cc(includes, "ocb.h",
+                                    "nettle_ocb_encrypt"))
+}
+
 #[cfg(target_env = "msvc")]
 fn try_vcpkg() -> Result<Config> {
     let lib = vcpkg::Config::new()
@@ -95,57 +121,35 @@ fn try_vcpkg() -> Result<Config> {
         .find_package("nettle")?;
 
     Ok(Config {
-        have_cv448: check_cv448(&include_paths),
-        include_paths,
+        have_cv448: check_cv448(&lib.include_paths),
+        have_ocb: check_ocb(&include_paths),
+        include_paths: lib.include_paths,
     })
 }
 
 #[cfg(not(target_env = "msvc"))]
 fn try_vcpkg() -> Result<Config> { Err("not applicable")?; unreachable!() }
 
-fn print_library(lib: &pkg_config::Library, mode: &str) {
-	for p in &lib.include_paths {
-		println!("cargo:include={}", p.display());
-	}
-
-	for p in &lib.frameworks {
-		println!("cargo:rustc-link-lib=framework={}", p);
-	}
-
-	for p in &lib.framework_paths {
-		println!("cargo:rustc-link-search=framework={}", p.display());
-	}
-
-    for p in &lib.libs {
-        println!("cargo:rustc-link-lib={}={}", mode, p);
-    }
-
-    for p in &lib.link_paths {
-        println!("cargo:rustc-link-search=native={}", p.display());
-    }
-}
-
 fn try_pkg_config() -> Result<Config> {
-    let mut nettle = pkg_config::probe_library("nettle hogweed")?;
-
-    let mode = match env::var_os("NETTLE_STATIC") {
-        Some(_) => "static",
-        None => "dylib",
-    };
-
-    print_library(&nettle, mode);
+    let mut nettle= Library::from(pkg_config::probe_library("nettle")?)
+        .merge(pkg_config::probe_library("hogweed")?.into());
 
     // GMP only got pkg-config support in release 6.2 (2020-01-18). So we'll try to use it if
     // possible, but fall back to just looking for it in the default include and lib paths.
-    if let Ok(mut gmp) = pkg_config::probe_library("gmp") {
-        print_library(&gmp, mode);
-        nettle.include_paths.append(&mut gmp.include_paths);
+    if let Ok(gmp) = pkg_config::probe_library("gmp").map(Into::into) {
+        nettle = nettle.merge(gmp);
     } else {
+        let mode = match env_var("GMP_STATIC").ok() {
+            Some(_) => "static",
+            None => "dylib",
+        };
+
         println!("cargo:rustc-link-lib={}=gmp", mode);
     }
 
     Ok(Config {
         have_cv448: check_cv448(&nettle.include_paths),
+        have_ocb: check_ocb(&nettle.include_paths),
         include_paths: nettle.include_paths,
     })
 }
@@ -153,15 +157,24 @@ fn try_pkg_config() -> Result<Config> {
 const NETTLE_PREGENERATED_BINDINGS: &str = "NETTLE_PREGENERATED_BINDINGS";
 
 fn real_main() -> Result<()> {
-    println!("cargo:rerun-if-env-changed=NETTLE_STATIC");
-    println!("cargo:rerun-if-env-changed={}", NETTLE_PREGENERATED_BINDINGS);
-
     let config = try_vcpkg().or_else(|_| try_pkg_config())?;
 
-    let out_path = Path::new(&env::var("OUT_DIR").unwrap()).join("bindings.rs");
+    // There is a bug (or obscure feature) in pkg-config-rs that
+    // doesn't let you statically link against libraries installed in
+    // /usr: https://github.com/rust-lang/pkg-config-rs/issues/102
+    //
+    // Work around that by emitting the directives necessary for cargo
+    // to statically link against Nettle, Hogweed, and GMP.
+    if env_var("NETTLE_STATIC").is_ok() {
+        println!("cargo:rustc-link-lib=static=nettle");
+        println!("cargo:rustc-link-lib=static=hogweed");
+        println!("cargo:rustc-link-lib=static=gmp");
+    }
+
+    let out_path = Path::new(&env_var("OUT_DIR").unwrap()).join("bindings.rs");
 
     // Check if we have a bundled bindings.rs.
-    if let Ok(bundled) = env::var(NETTLE_PREGENERATED_BINDINGS)
+    if let Ok(bundled) = env_var(NETTLE_PREGENERATED_BINDINGS)
     {
         let p = Path::new(&bundled);
         if p.exists() {
@@ -177,19 +190,23 @@ fn real_main() -> Result<()> {
         // Includes all nettle headers except mini-gmp.h
         .header("bindgen-wrapper.h")
         // Workaround for https://github.com/rust-lang-nursery/rust-bindgen/issues/550
-        .blacklist_type("max_align_t")
-        .blacklist_function("strtold")
-        .blacklist_function("qecvt")
-        .blacklist_function("qfcvt")
-        .blacklist_function("qgcvt")
-        .blacklist_function("qecvt_r")
-        .blacklist_function("qfcvt_r")
+        .blocklist_type("max_align_t")
+        .blocklist_function("strtold")
+        .blocklist_function("qecvt")
+        .blocklist_function("qfcvt")
+        .blocklist_function("qgcvt")
+        .blocklist_function("qecvt_r")
+        .blocklist_function("qfcvt_r")
         .size_t_is_usize(true);
 
     if config.have_cv448 {
         builder = builder.header_contents("cv448-wrapper.h", "#include <nettle/curve448.h>");
     }
 
+    if config.have_ocb {
+        builder = builder.header_contents("ocb-wrapper.h", "#include <nettle/ocb.h>");
+    }
+
     for p in config.include_paths {
         builder = builder.clang_arg(format!("-I{}", p.display()));
     }
@@ -203,14 +220,25 @@ fn real_main() -> Result<()> {
 pub mod config {{
     /// Cv448/Ed448 support in Nettle.
     pub const HAVE_CV448: bool = {};
+
+    /// OCB support in Nettle.
+    pub const HAVE_OCB: bool = {};
 }}
-", config.have_cv448)?;
+",
+             config.have_cv448,
+             config.have_ocb,
+    )?;
 
     if ! config.have_cv448 {
         let mut stubs = fs::File::open("cv448-stubs.rs")?;
         io::copy(&mut stubs, &mut s)?;
     }
 
+    if ! config.have_ocb {
+        let mut stubs = fs::File::open("ocb-stubs.rs")?;
+        io::copy(&mut stubs, &mut s)?;
+    }
+
     Ok(())
 }
 
@@ -235,7 +263,7 @@ fn print_hints() {
     if cfg!(target_os = "windows") {
         eprintln!("If you are using MSYS2, try:
 
-    $ pacboy -S clang:x nettle:x
+    $ pacman -S mingw-w64-x86_64-{{clang,pkg-config,nettle}} libnettle-devel
 ");
     } else if cfg!(target_os = "macos") {
         eprintln!("If you are using MacPorts, try:
@@ -272,3 +300,81 @@ fn log<S: AsRef<str>>(m: S) {
     writeln!(&mut fs::OpenOptions::new().append(true).open("/tmp/l").unwrap(),
              "{}", m.as_ref()).unwrap();
 }
+
+/// Somewhat like pkg_config::Library, but only with the parts we use.
+#[derive(Default)]
+pub struct Library {
+    pub libs: Vec<String>,
+    pub link_paths: Vec<PathBuf>,
+    pub frameworks: Vec<String>,
+    pub framework_paths: Vec<PathBuf>,
+    pub include_paths: Vec<PathBuf>,
+}
+
+impl Library {
+    /// Merges two libraries into one.
+    fn merge(mut self, mut other: Self) -> Self {
+        self.libs.append(&mut other.libs);
+        self.libs.sort();
+        self.libs.dedup();
+
+        self.link_paths.append(&mut other.link_paths);
+        self.link_paths.sort();
+        self.link_paths.dedup();
+
+        self.frameworks.append(&mut other.frameworks);
+        self.frameworks.sort();
+        self.frameworks.dedup();
+
+        self.framework_paths.append(&mut other.framework_paths);
+        self.framework_paths.sort();
+        self.framework_paths.dedup();
+
+        self.include_paths.append(&mut other.include_paths);
+        self.include_paths.sort();
+        self.include_paths.dedup();
+
+        self
+    }
+
+    /// Emits directives to make cargo link to the library.
+    ///
+    /// Note: If we are using pkg_config, then the necessary
+    /// directives are emitted by that crate already.  Emitting these
+    /// ourselves may be useful in the future if we add other ways to
+    /// configure how to build against Nettle.
+    #[allow(dead_code)]
+    fn print_library(&self, mode: &str) {
+	for p in &self.include_paths {
+	    println!("cargo:include={}", p.display());
+	}
+
+	for p in &self.frameworks {
+	    println!("cargo:rustc-link-lib=framework={}", p);
+	}
+
+	for p in &self.framework_paths {
+	    println!("cargo:rustc-link-search=framework={}", p.display());
+	}
+
+        for p in &self.libs {
+            println!("cargo:rustc-link-lib={}={}", mode, p);
+        }
+
+        for p in &self.link_paths {
+            println!("cargo:rustc-link-search=native={}", p.display());
+        }
+    }
+}
+
+impl From<pkg_config::Library> for Library {
+    fn from(l: pkg_config::Library) -> Library {
+        Library {
+            libs: l.libs,
+            link_paths: l.link_paths,
+            frameworks: l.frameworks,
+            framework_paths: l.framework_paths,
+            include_paths: l.include_paths,
+        }
+    }
+}
